# -*- coding: utf-8 -*-
##############################################################################
#
#    ebicspy, EBICS protocol library
#    Copyright (C) 2013-2014 Aurélien DUMAINE (aurelien.dumaine@free.fr).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from Crypto.PublicKey import RSA
import hashlib
from crypto_tools import *
from ini_step import *
from hia_step import *
from hpb_step import *

class Bank :
    def __init__(self, storageService, name, urlHost, urlPort, urlRoot, hostId, keysDirectory="keys") :
        self.name = name
        self.urlHost = urlHost
        self.urlPort = urlPort
        self.urlRoot = urlRoot
        self.hostId = hostId
#        self.keysDirectory = keysDirectory
        self.storageService = storageService


    def getKeysDirectory (self) :
        return self.keysDirectory

    def getHostId (self) :
        return self.hostId

    def getUrlHost (self) :
        return self.urlHost

    def getUrlPort (self) :
        return self.urlPort

    def getUrlRoot (self) :
        return self.urlRoot

    def import_TLSCertificate(self, path):
        #FIXME : check the TLS certificate each time we send/receive data from/to the bank
        print "TODO"

    def setAuthKey (self, authKey) :
        self.authKey = authKey

    def getKeyHash(self, bankKey):
        #The SHA-256 hash values of the financial institution's public keys for X002 and
        #E002 are composed by concatenating the exponent with a blank character and the
        #modulus in hexadecimal representation (using lower case letters) without leading
        #zero (as to the hexadecimal representation). The resulting string has to be
        #converted into a byte array based on US ASCII code

        n_long = getattr(bankKey.key, 'n')
        n_hex_str = str(hex(n_long))[2:-1]

        e_long = getattr(bankKey.key, 'e')
        e_hex_str = str(hex(e_long))[2:-1]

        s = e_hex_str + ' ' + n_hex_str
        if s[0] == '0':
            s = s[1:]

        m = hashlib.sha256()
        m.update(s)
        string_digest = m.digest()

        res = string_digest.encode('base64')
        return res.strip()
        
    def saveKeys(self, auth_modulus, auth_exponent, encrypt_modulus, encrypt_exponent):
        self.storageService.saveBankKeys(auth_modulus, auth_exponent, encrypt_modulus, encrypt_exponent)

    def getAuthKey (self) :
        return self.authKey

    def setSignKey (self, signKey) :
        self.signKey = signKey

    def getSignKey (self) :
        return self.signKey

    def setEncryptKey (self, encryptKey) :
        self.encryptKey = encryptKey

    def getEncryptKey (self) :
        return self.encryptKey

    def getName (self) :
        return self.name

class Partner : 
    def __init__(self, storageService, name, partnerId, userId, fileExchangedLogger, keysDirectory="keys", ebicsProfile="T", signCertificatePath=""):
        self.name = name
        self.partnerId = partnerId
        self.userId = userId
#        self.keysDirectory = keysDirectory
        self.ebicsProfile = ebicsProfile
        self.signCertificatePath = signCertificatePath
        self.fileExchangedLogger = fileExchangedLogger
        self.storageService = storageService

    def handle_ini_exchange(self, bank):
        ini_exchange(self, bank)
        #TODO check the validity of the response
        ini_sign_letter = letter("INI",
            "Certificat pour la signature électronique (A005)" , 
            partner.getSignCertificate(),
            bank.getHostId(), 
            partner.getUserId(), 
            partner.getPartnerId())
        print ini_sign_letter
        self.storageService.saveLetter(ini_sign_letter, "INI_letter_A005", partner.getName(), bank.getName())
        print "========== ENVOI MESSAGE INI TERMINE =========="
        print "========== IL FAUT MAINTENANT ENVOYER LE MESSAGE HIA =========="

    def handle_hia_exchange(self, bank):
        hia_exchange(self, bank)
        #TODO check the validity of the response
        hia_encrypt_letter = letter("HIA",
            "Certificat pour le chiffrement (E002)" , 
            partner.getEncryptCertificate(),
            bank.getHostId(), 
            partner.getUserId(), 
            partner.getPartnerId())
        print hia_encrypt_letter
        self.storageService.saveLetter(hia_encrypt_letter, "HIA_letter_E002_", partner.getName(), bank.getName())
        hia_auth_letter = letter("HIA",
            "Certificat pour l'authentification (X002)" , 
            partner.getAuthCertificate(),
            bank.getHostId(), 
            partner.getUserId(), 
            partner.getPartnerId())
        print hia_auth_letter
        self.storageService.saveLetter(hia_auth_letter, "HIA_letter_X002_", partner.getName(), bank.getName())
        print "========== ENVOI MESSAGE HIA TERMINE =========="
        print "===>>> VOUS DEVEZ ENVOYER VOS LETTRES D'INITIALISATION AVANT DE RECUPERER LES CLES DE LA BANQUE"
        #TODO display the ini letter file path
    

    def init(self, bank):
        partner_keys_already_exists = self.storageService.partner_keys_already_exists(self.getName())

        if partner_keys_already_exists == False:
            self.createKeys()
            print "========== CREATION DES CLES TERMINNEE =========="
            sign_cert = self.storageService.loadCertificate('sign', self.getName())
            self.setSignCertificate(sign_cert)
            auth_cert = self.storageService.loadCertificate('auth', self.getName())
            self.setAuthCertificate(auth_cert)
            encrypt_cert = self.storageService.loadCertificate('encrypt', self.getName())
            self.setEncryptCertificate(encrypt_cert)
            
            self.handle_ini_exchange(bank)
            self.handle_ini_exchange(bank)
            
        else :
            self.loadKeys()
                            
            sign_cert = self.storageService.loadCertificate('sign', self.getName())
            self.setSignCertificate(sign_cert)
            auth_cert = self.storageService.loadCertificate('auth', self.getName())
            self.setAuthCertificate(auth_cert)
            encrypt_cert = self.storageService.loadCertificate('encrypt', self.getName())
            self.setEncryptCertificate(encrypt_cert)
            print "========== CHARGEMENT DES CLES ET DES CERTIFICATS PARTNER TERMINE =========="


            #FIXME : get auth and encrypt version using HPD request
            auth_version = 'X002'
            encrypt_version = 'E002'
                    
            bank_keys_already_exists = self.storageService.bank_keys_already_exists(bank.getName(), auth_version)
            
            if bank_keys_already_exists == False :
                bank_auth_key_hash = self.storageService.getBankAuthKeyHash()
                bank_encrypt_key_hash = self.storageService.getBankEncryptKeyHash()
                hpb_exchange(self, bank, bank_auth_key_hash, bank_encrypt_key_hash)

            else :
                auth_modulus = self.storageService.getBankKeyComponent(bank.getName(), "modulus", auth_version)
                auth_exponent = self.storageService.getBankKeyComponent(bank.getName(), "public_exponent", auth_version)
                bank.setAuthKey(RSA.construct((auth_modulus, auth_exponent)))
                encrypt_modulus = self.storageService.getBankKeyComponent(bank.getName(), "modulus", encrypt_version)
                encrypt_exponent = self.storageService.getBankKeyComponent(bank.getName(), "public_exponent", encrypt_version)                
                bank.setEncryptKey(RSA.construct((encrypt_modulus, encrypt_exponent)))
                print "========== CHARGEMENT DES CLES ET DES CERTIFICATS DE LA BANQUE TERMINE =========="

    def getEbicsProfile (self) :
        return self.ebicsProfile

    def getFileExchangedLogger (self) :
        return self.fileExchangedLogger
        
    def getName (self) :
        return self.name

#    def getKeysDirectory (self) :
#        return self.keysDirectory

    def getPartnerId (self) :
        return self.partnerId

    def getUserId (self) :
        return self.userId

    def setAuthKey (self, private, public) :
        self.authPrivateKey= private
        self.authPublicKey = public
        #print "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM"
        #print getattr(self.getAuthPrivateKey().key, 'n')
        #print getattr(self.getAuthPrivateKey().key, 'd')
        #print getattr(self.getAuthPrivateKey().key, 'e')

    def getAuthPrivateKey (self) :
        return self.authPrivateKey

    def getAuthPublicKey (self) :
        return self.authPublicKey

    def setSignKey (self, private, public) :
        self.signPrivateKey = private
        self.signPublicKey = public

    def getSignPrivateKey (self) :
        return self.signPrivateKey

    def getSignPublicKey (self) :
        return self.signPublicKey

    def setEncryptKey (self, private, public) :
        self.encryptPrivateKey = private
        self.encryptPublicKey = public

    def getEncryptPrivateKey (self) :
        return self.encryptPrivateKey

    def getEncryptPublicKey (self) :
        return self.encryptPublicKey

    def setSignCertificate (self, certificate):
        self.signCertificate = certificate

    def getSignCertificate (self):
        return self.signCertificate

    def setAuthCertificate (self, certificate):
        self.authCertificate = certificate

    def getAuthCertificate (self):
        return self.authCertificate

    def setEncryptCertificate (self, certificate):
        self.encryptCertificate = certificate

    def getEncryptCertificate (self):
        return self.encryptCertificate

    def createKeys(self):
        auth_private, auth_public, auth_cert = create_rsa_key(AUTH_KEY_LENGTH, self.name, "auth", self.storageService)
        self.setAuthKey(auth_private, auth_public)

        if self.ebicsProfile == "T" :
            sign_private, sign_public, sign_cert = create_rsa_key(SIGN_KEY_LENGTH, self.name, "sign", self.storageService) 
            self.setSignKey(sign_private, sign_public)

        encrypt_private, encrypt_public, encrypt_cert = create_rsa_key(ENCRYPT_KEY_LENGTH, self.name, "encrypt", self.storageService) 
        self.setEncryptKey(encrypt_private, encrypt_public)
#        print("=================================================")
#        print getattr(auth_private.key, 'n')
#        print getattr(auth_private.key, 'd')
#        print getattr(auth_private.key, 'e')
        self.loadKeys()
#        print "OK"
#        print("=================================================")
#        print getattr(auth_private.key, 'n')
#        print getattr(auth_private.key, 'd')
#        print getattr(auth_private.key, 'e')

    def loadKeys(self, path_auth_publicKey, path_auth_privateKey, path_sign_publicKey, path_sign_privateKey, path_encrypt_publicKey, path_encrypt_privateKey):
        privateKey = self.storageService.getPartnerKey(self.getName(), 'auth', 'private')
        publicKey = self.storageService.getPartnerKey(self.getName(), 'auth', 'public')
        self.setAuthKey(RSA.importKey(privateKey), RSA.importKey(publicKey))

        if self.ebicsProfile == "TS" :
            privateKey = open(self.signCertificatePath, "r").read()
            #FIXME : does RSA.import key transform the privateKeyin public key object readng the privateKeyFile ?
            publicKey = open(self.signCertificatePath, "r").read()
            self.setSignKey(RSA.importKey(privateKey), RSA.importKey(publicKey))
        else :
            privateKey = self.storageService.getPartnerKey(self.getName(), 'sign', 'private')
            publicKey = self.storageService.getPartnerKey(self.getName(), 'sign', 'public')
            self.setSignKey(RSA.importKey(privateKey), RSA.importKey(publicKey))

        privateKey = self.storageService.getPartnerKey(self.getName(), 'encrypt', 'private')
        publicKey = self.storageService.getPartnerKey(self.getName(), 'encrypt', 'public')
        self.setEncryptKey(RSA.importKey(privateKey), RSA.importKey(publicKey))


# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4
